# tests.status.test_aggregator
#
# Author: Alejandro J. Cura <alecu@canonical.com>
#
# Copyright 2011 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Tests for the status events aggregator."""

from twisted.internet.task import Clock
from twisted.trial.unittest import TestCase

from ubuntuone.status import aggregator
from ubuntuone.syncdaemon.status_listener import StatusListener


class FakeNotification(object):
    """A fake notification class."""

    def __init__(self, application_name="fake app"):
        """Initialize this instance."""
        self.notifications_shown = []
        self.application_name = application_name

    def send_notification(self, title, message, icon=None, append=False):
        """Show a notification to the user."""
        self.notifications_shown.append((title, message, icon, append))


class DelayedBufferTestCase(TestCase):
    """Test the delayed buffer class."""

    THRESHOLD = 5.0
    SMALL_DELAY = 0.5
    COUNT_MESSAGES = 4

    def setUp(self):
        """Initialize this test instance."""
        self.result = []
        self.clock = Clock()
        append = lambda *args: self.result.append(args)
        self.ag = aggregator.DelayedBuffer(cb=append, clock=self.clock,
                                        threshold=self.THRESHOLD)

    def test_a_bunch_at_once_aggregated_as_one(self):
        """A bunch of messages sent at once are aggregated into one."""
        for n in range(self.COUNT_MESSAGES):
            self.ag.push("message1")
        self.clock.advance(self.THRESHOLD)
        self.assertEqual(len(self.result), 1)

    def test_a_bunch_under_threshold_aggregated_as_one(self):
        """A bunch of messages sent under threshold are aggregated into one."""
        for n in range(self.COUNT_MESSAGES):
            self.ag.push("message1")
            self.clock.advance(self.SMALL_DELAY)
        self.clock.advance(self.THRESHOLD)
        self.assertEqual(len(self.result), 1)

    def test_a_bunch_over_threshold_are_not_aggregated(self):
        """A bunch of messages sent over the threshold are not aggregated."""
        for n in range(self.COUNT_MESSAGES):
            self.ag.push("message1")
            self.clock.advance(self.THRESHOLD)
        self.clock.advance(self.THRESHOLD + self.SMALL_DELAY)
        self.assertEqual(len(self.result), self.COUNT_MESSAGES)


class FakeAggregator(object):
    """Appends all status pushed into a list."""

    def __init__(self, *args, **kwargs):
        """Initialize this instance."""
        self.statuses = []

    def push(self, event):
        self.statuses.append(event)


class StatusAggregatorFrontendTestCase(TestCase):
    """Test the status events aggregator."""

    def setUp(self):
        """Initialize this test instance."""
        self.fakefsm = None
        self.fakevm = None
        self.patch(aggregator, "DelayedBuffer", FakeAggregator)
        self.status_aggregator = aggregator.StatusAggregator()
        self.listener = StatusListener(self.fakefsm, self.fakevm,
                                       self.status_aggregator)

    def test_file_published(self):
        """A file published event is processed."""
        share_id = "fake share id"
        node_id = "fake node id"
        is_public = True
        public_url = "http://fake_public/url"

        self.listener.handle_AQ_CHANGE_PUBLIC_ACCESS_OK(share_id, node_id,
                                                        is_public, public_url)
        status = self.status_aggregator.aggregator.statuses[0]
        self.assertIsInstance(status, aggregator.FilePublishingStatus)

    def test_file_unpublished(self):
        """A file unpublished event is processed."""
        share_id = "fake share id"
        node_id = "fake node id"
        is_public = False
        public_url = None # SD sends None when unpublishing

        self.listener.handle_AQ_CHANGE_PUBLIC_ACCESS_OK(share_id, node_id,
                                                        is_public, public_url)
        status = self.status_aggregator.aggregator.statuses[0]
        self.assertIsInstance(status, aggregator.FileUnpublishingStatus)


class FakeOtherStatus(aggregator.StatusEvent):
    """A fake status to test weight comparisons."""

    def __init__(self, weight):
        """Initialize with the fake weight."""
        super(FakeOtherStatus, self).__init__()
        self.WEIGHT = weight

class StatusEventTestCase(TestCase):

    """Test the status event class and children."""

    CLASS = aggregator.StatusEvent
    CLASS_ARGS = ()
    status = None

    def setUp(self):
        """Initialize this test instance."""
        if type(self) == StatusEventTestCase:
            self.assertRaises(AssertionError, self.CLASS, self.CLASS_ARGS)
        else:
            self.status = self.CLASS(self.CLASS_ARGS)

    def test_one_message_built_correctly(self):
        """The message returned by one() is returned ok."""
        if self.status:
            self.assertEqual(self.status.one(), self.CLASS.MESSAGE_ONE)

    def test_many_message_built_correctly(self):
        """The message returned by many() is returned ok."""
        if self.status:
            test_count = 99
            expected = self.CLASS.MESSAGE_MANY % test_count
            self.assertEqual(self.status.many(test_count), expected)

    def test_compare_sorts_according_to_weight(self):
        if self.status:
            other = FakeOtherStatus(99)
            self.assertTrue(cmp(self.status, other) < 0)
            self.assertTrue(cmp(other, self.status) > 0)


class FilePublishingStatusTestCase(StatusEventTestCase):
    """Test the file publishing status class."""

    CLASS = aggregator.FilePublishingStatus
    CLASS_ARGS = ("http://fake_public/url",)

    def test_one_message_built_correctly(self):
        """The message returned by one() should include the url."""
        expected = self.CLASS.MESSAGE_ONE % self.status.args
        self.assertEqual(self.status.one(), expected)


class FileUnpublishingStatusTestCase(StatusEventTestCase):
    """Test the file unpublishing status class."""

    CLASS = aggregator.FileUnpublishingStatus
    CLASS_ARGS = (None,)


class StatusAggregatorBackendTestCase(TestCase):
    """Test the callback of the status aggregator."""

    def setUp(self):
        """Initialize this test instance."""
        self.patch(aggregator, "Notification", FakeNotification)
        self.patch(aggregator, "DelayedBuffer", FakeAggregator)
        self.status_aggregator = aggregator.StatusAggregator()
        self.fake_notification = self.status_aggregator.notification

    def test_pushed_one(self):
        """One event pushed."""
        public_url = "http://fake_public/url"
        event = aggregator.FilePublishingStatus(public_url)
        events = [event]
        self.status_aggregator.callback(events)
        message = aggregator.FilePublishingStatus.MESSAGE_ONE % public_url
        notification = (aggregator.UBUNTUONE_TITLE, message, None, False)
        self.assertIn(notification, self.fake_notification.notifications_shown)

    def test_pushed_many_same_type(self):
        """Pushed many events, all of the same type."""
        public_url = "http://fake_public/url%d"
        count = 10
        events = []

        for n in range(count):
            event = aggregator.FilePublishingStatus(public_url % n)
            events.append(event)

        self.status_aggregator.callback(events)
        message = aggregator.FilePublishingStatus.MESSAGE_MANY % count
        notification = (aggregator.UBUNTUONE_TITLE, message, None, False)
        self.assertIn(notification, self.fake_notification.notifications_shown)

    def test_pushed_many_mixed_types(self):
        """Pushed many events, of mixed types."""
        public_url = "http://fake_public/url%d"
        count = 5
        events = []

        for n in range(count):
            event = aggregator.FilePublishingStatus(public_url % n)
            events.append(event)
            event = aggregator.FileUnpublishingStatus(public_url % n)
            events.append(event)

        self.status_aggregator.callback(events)
        lines = [aggregator.FilePublishingStatus.MESSAGE_MANY % count,
                 aggregator.FileUnpublishingStatus.MESSAGE_MANY % count]
        message = "\n".join(lines)
        notification = (aggregator.UBUNTUONE_TITLE, message, None, False)
        self.assertIn(notification, self.fake_notification.notifications_shown)
