# ubuntuone.status.aggregator
#
# Author: Alejandro J. Cura <alecu@canonical.com>
#
# Copyright 2011 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Aggregate status events."""

import itertools

from twisted.internet import reactor

from ubuntuone.platform.notification import Notification

UBUNTUONE_TITLE = "Ubuntu One"

class StatusEvent(object):
    """An event representing a status change."""

    MESSAGE_ONE = None # to be defined in child classes
    MESSAGE_MANY = None # to be defined in child classes
    WEIGHT = 99
    DO_NOT_INSTANCE = "Do not instance this class, only children."""

    def __init__(self, *args):
        """Initialize this instance."""
        assert type(self) != StatusEvent, self.DO_NOT_INSTANCE
        self.args = args

    def __cmp__(self, other):
        """These events are displayed sorted by their weight."""
        assert isinstance(other, StatusEvent)
        return self.WEIGHT - other.WEIGHT

    def one(self):
        """A message if this is the only event of this type."""
        return self.MESSAGE_ONE

    def many(self, count):
        """A message if there are many events of this type."""
        return self.MESSAGE_MANY % count


class FilePublishingStatus(StatusEvent):
    """Files that are made public with a url."""

    MESSAGE_ONE = "A file was just made public at %s"
    MESSAGE_MANY = "%d files were just made public"
    WEIGHT = 50

    def one(self):
        """Show the url if only one event of this type."""
        return self.MESSAGE_ONE % self.args


class FileUnpublishingStatus(StatusEvent):
    """Files that have stopped being published."""

    MESSAGE_ONE = "A file is no longer published"
    MESSAGE_MANY = "%d files are no longer published"
    WEIGHT = 51


class DelayedBuffer(object):
    """Buffers events received closely in time."""

    def __init__(self, cb, threshold=5, clock=reactor):
        """Initialize this instance with the callback, threshold and clock."""
        self.pending = []
        self.cb = cb
        self.clock = clock
        self.threshold = threshold
        self.delayedcall = None

    def push(self, event):
        """Push a new event into this aggregator."""
        self.pending.append(event)
        if self.delayedcall and self.delayedcall.active():
            self.delayedcall.reset(self.threshold)
        else:
            self.delayedcall = self.clock.callLater(self.threshold,
                                                    self.handle_timeout)

    def handle_timeout(self, *args):
        """A timeout has happened, so aggregate pending events."""
        self.cb(self.pending)
        self.pending = []


class StatusAggregator(object):
    """Aggregates status events."""

    def __init__(self, clock=reactor):
        """Initialize this instance."""
        self.notification = Notification()
        self.aggregator = DelayedBuffer(self.callback, threshold=5,
                                              clock=clock)

    def group_statuses(self, status_events):
        """Groups statuses by type and weight."""
        return itertools.groupby(sorted(status_events), type)

    def callback(self, status_events):
        """Pushed a bunch of events by the aggregator."""

        lines = []
        for klass, statuses in self.group_statuses(status_events):
            statuses = list(statuses)
            count = len(statuses)
            first_status = statuses[0]
            if count == 1:
                lines.append(first_status.one())
            else:
                lines.append(first_status.many(count))

        text = "\n".join(lines)
        self.notification.send_notification(UBUNTUONE_TITLE, text)

    def file_published(self, public_url):
        """A file was published."""
        self.aggregator.push(FilePublishingStatus(public_url))

    def file_unpublished(self, public_url):
        """A file was unpublished."""
        self.aggregator.push(FileUnpublishingStatus(public_url))
